import pygame
import typing
from mili import error as _error
from dataclasses import dataclass

if typing.TYPE_CHECKING:
    from mili import MILI as _MILI

__all__ = ("ImageCache", "Interaction", "ElementData")


class ImageCache:
    _preallocated_caches = []
    _preallocated_index = -1

    def __init__(self):
        self._cache = None

    def get_output(self) -> pygame.Surface:
        if self._cache is None:
            return None
        return self._cache.get("output", None)

    @classmethod
    def preallocate_caches(cls, amount: int):
        cls._preallocated_caches = [ImageCache() for _ in range(amount)]

    @classmethod
    def get_next_cache(cls):
        cls._preallocated_index += 1
        if cls._preallocated_index > len(cls._preallocated_caches) - 1:
            raise _error.MILIStatusError(
                "Too few image caches preallocated, failed to retrieve the next one"
            )
        return cls._preallocated_caches[cls._preallocated_index]


@dataclass
class Interaction:
    _mili: "_MILI"
    id: int
    hovered: bool
    press_button: int
    just_pressed_button: int
    just_released_button: int
    absolute_hover: bool
    unhover_pressed: bool
    just_hovered: bool
    just_unhovered: bool

    @property
    def left_pressed(self) -> bool:
        return self.press_button == pygame.BUTTON_LEFT

    @property
    def left_just_pressed(self) -> bool:
        return self.just_pressed_button == pygame.BUTTON_LEFT

    @property
    def left_just_released(self) -> bool:
        return self.just_released_button == pygame.BUTTON_LEFT

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        self._mili.end()


class ElementGridData:
    overflowx: float
    overflowy: float
    padx: float
    pady: float
    spacing: float

    def __init__(self, **kwargs):
        for name, value in kwargs.items():
            setattr(self, name, value)


class ElementData:
    interaction: Interaction
    rect: pygame.Rect
    absolute_rect: pygame.Rect
    z: int
    id: int
    style: dict[str, typing.Any]
    children_ids: list[int]
    components: dict[typing.Literal["type", "style", "data"], typing.Any]
    parent_id: int
    grid: ElementGridData | None

    def __init__(self, mili: "_MILI", **kwargs):
        self._mili = mili
        for name, value in kwargs.items():
            setattr(self, name, value)

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        self._mili.end()

    def __getattr__(self, name):
        if hasattr(self.interaction, name):
            return getattr(self.interaction, name)
        raise AttributeError



from .mili import MILI, register_custom_component, pack_component
from .data import ElementData, Interaction, ImageCache
from .style import RESIZE, X, CENTER, PADLESS, FLOATING
from .utility import (
    Selectable,
    Dragger,
    Scroll,
    Scrollbar,
    Slider,
    GenericApp,
    InteractionSound,
    CustomWindowBorders,
    percentage,
    indent,
    fit_image,
)

from . import error
from . import style
from . import typing
from . import data
from . import animation

from collections import namedtuple

VERSION = namedtuple("MILIVersionInfo", "major minor micro")(0, 9, 6)
VERSION_STR = "0.9.6"
del namedtuple

__all__ = (
    "MILI",
    "ElementData",
    "Interaction",
    "ImageCache",
    "Selectable",
    "Dragger",
    "Scroll",
    "Scrollbar",
    "Slider",
    "GenericApp",
    "InteractionSound",
    "CustomWindowBorders",
    "percentage",
    "indent",
    "fit_image",
    "register_custom_component",
    "pack_component",
    "RESIZE",
    "X",
    "CENTER",
    "PADLESS",
    "FLOATING",
    "error",
    "style",
    "typing",
    "data",
    "animation",
    "VERSION",
    "VERSION_STR",
)



import sys
import webbrowser


def main():
    sys.stdout.write("\033[F")
    sys.stdout.write("\033[K")
    sys.stdout.flush()

    if len(sys.argv) <= 1:
        print("ERROR: missing command 'guide'")
        return
    if len(sys.argv) > 2:
        print("ERROR: too many commands")
        return
    if sys.argv[1] != "guide":
        print(f"ERROR: wrong command '{sys.argv[1]}', expected 'guide'")
        return

    url = "https://github.com/damusss/mili/blob/main/guide/guide.md"
    print(f"Opening guide at '{url}'")
    webbrowser.open(url)


if __name__ == "__main__":
    main()



import pygame
import typing
from mili import error
from mili import data as _data
from mili import typing as _typing

if typing.TYPE_CHECKING:
    from mili import MILI as _MILI

__all__ = ()


class _globalctx:
    _font_cache = {}
    _component_types: dict[str, str | _typing.ComponentProtocol] = dict.fromkeys(
        ["rect", "circle", "line", "polygon", "line", "text", "image"], "builtin"
    )
    _mili_stack = []
    _mili: "_MILI" = None

    @staticmethod
    def _abs_perc(val, dim) -> float:
        if isinstance(val, str):
            sign = 1
            if val.startswith("-"):
                val = val.replace("-", "")
                sign = -1
            try:
                val = float(val)
            except ValueError:
                raise error.MILIValueError(f"Invalid percentage value '{val}'")
            return ((dim * float(val)) / 100) * sign
        return val

    @staticmethod
    def _get_first_button(buttons):
        bid = -1
        for i, b in enumerate(buttons):
            if b:
                bid = i + 1
                break
        return bid

    @staticmethod
    def _nine_patch(image: pygame.Surface, w, h, s, smoothscale):
        ow, oh = image.get_size()
        scale_func = (
            pygame.transform.smoothscale if smoothscale else pygame.transform.scale
        )
        subsurf = image.subsurface
        ohsize, hsize = ow - s * 2, w - s * 2
        ovsize, vsize = oh - s * 2, h - s * 2
        base = pygame.Surface((w, h), pygame.SRCALPHA)
        base.blit(subsurf((0, 0, s, s)), (0, 0))  # topleft
        base.blit(subsurf((ow - s, 0, s, s)), (w - s, 0))  # topright
        base.blit(subsurf((0, oh - s, s, s)), (0, h - s))  # bottomleft
        base.blit(subsurf((ow - s, oh - s, s, s)), (w - s, h - s))  # bottomright
        if ohsize > 0:
            base.blit(scale_func(subsurf((s, 0, ohsize, s)), (hsize, s)), (s, 0))  # top
            base.blit(
                scale_func(subsurf((s, oh - s, ohsize, s)), (hsize, s)),
                (s, h - s),
            )  # bottom
        if ovsize > 0:
            base.blit(
                scale_func(subsurf((0, s, s, ovsize)), (s, vsize)), (0, s)
            )  # left
            base.blit(
                scale_func(subsurf((ow - s, s, s, ovsize)), (s, vsize)),
                (w - s, s),
            )  # right
        if ohsize > 0 and ovsize > 0:
            base.blit(
                scale_func(subsurf((s, s, ohsize, ovsize)), (hsize, vsize)),
                (s, s),
            )  # inner
        return base

    @staticmethod
    def _get_image(
        rect: pygame.Rect,
        data: pygame.Surface,
        padx,
        pady,
        do_fill,
        do_stretchx,
        do_stretchy,
        fill_color,
        border_radius,
        alpha,
        smoothscale,
        nine_patch,
    ) -> pygame.Surface:
        iw, ih = data.size
        tw, th = max(rect.w - padx * 2, 1), max(rect.h - pady * 2, 1)

        if nine_patch == 0:
            if not do_fill:
                w = tw
                h = int(ih * (w / iw))
                if h > th:
                    h = th
                    w = int(iw * (h / ih))
                w, h = max(w, 1), max(h, 1)
                if w < tw and do_stretchx:
                    w = tw
                if h < th and do_stretchy:
                    h = th
                if smoothscale:
                    image = pygame.transform.smoothscale(data, (w, h))
                else:
                    image = pygame.transform.scale(data, (w, h))
            else:
                w = tw
                h = int(ih * (w / iw))
                if h < th:
                    h = th
                    w = int(iw * (h / ih))
                if smoothscale:
                    image = pygame.transform.smoothscale(data, (w, h))
                else:
                    image = pygame.transform.scale(data, (w, h))
                image = image.subsurface((w // 2 - tw // 2, h // 2 - th // 2, tw, th))
        else:
            image = _globalctx._nine_patch(data, tw, th, nine_patch, smoothscale)

        w, h = image.size
        if fill_color is not None:
            image.fill(fill_color)
        if border_radius > 0:
            mask_surf = pygame.Surface((w, h), pygame.SRCALPHA)
            mask_surf.fill(0)
            pygame.draw.rect(
                mask_surf, (255, 255, 255, 255), (0, 0, w, h), 0, int(border_radius)
            )
            image.blit(mask_surf, (0, 0), special_flags=pygame.BLEND_RGBA_MULT)
        if alpha != 255:
            image.set_alpha(alpha)
        return image

    @staticmethod
    def _element_data(
        mili: "_MILI", el: dict[str, typing.Any], interaction: "_data.Interaction"
    ) -> "_data.ElementData":
        if el["id"] in mili._ctx._old_data:
            old_data = mili._ctx._old_data[el["id"]]
        else:
            old_data = {
                "rect": pygame.Rect(),
                "abs_rect": pygame.Rect(),
                "children_ids": [],
                "components": [],
                "parent_id": 0,
                "grid": None,
            }
        grid = old_data["grid"]
        return _data.ElementData(
            mili,
            interaction=interaction,
            rect=old_data["rect"].copy(),
            absolute_rect=old_data["abs_rect"].copy(),
            z=el["z"],
            id=el["id"],
            style=el["style"].copy(),
            children_ids=list(old_data["children_ids"]),
            parent_id=old_data["parent_id"],
            components=old_data["components"].copy(),
            grid=(
                _data.ElementGridData(
                    overflowx=-1, overflowy=-1, padx=0, pady=0, spacing=0
                )
                if grid is None
                else _data.ElementGridData(
                    overflowx=grid["overflowx"],
                    overflowy=grid["overflowy"],
                    padx=grid["padx"],
                    pady=grid["pady"],
                    spacing=grid["spacing"],
                )
            ),
        )

    @staticmethod
    def _text_align(align, rect: pygame.Rect, padx, pady):
        x, y = 0, 0
        if align == "center":
            x, y = rect.centerx, rect.centery
        elif align == "topleft":
            x, y = rect.x + padx, rect.y + pady
        elif align == "bottomleft":
            x, y = rect.x + padx, rect.bottom - pady
        elif align == "bottomright":
            x, y = rect.right - padx, rect.bottom - pady
        elif align == "topright":
            x, y = rect.right - padx, rect.y + pady
        elif align == "midleft" or align == "left":
            align = "midleft"
            x, y = rect.x + padx, rect.centery
        elif align == "midright" or align == "right":
            align = "midright"
            x, y = rect.right - padx, rect.centery
        elif align == "midtop" or align == "top":
            align = "midtop"
            x, y = rect.centerx, rect.y + pady
        elif align == "midbottom" or align == "bottom":
            align = "midbottom"
            x, y = rect.centerx, rect.bottom - pady
        else:
            raise error.MILIValueError("Invalid text alignment")
        return {align: (x, y)}


class _ctx:
    def __init__(self, mili: "_MILI"):
        self._mili = mili
        self._parent: dict[str, typing.Any] = {
            "rect": pygame.Rect(0, 0, 1, 1),
            "abs_rect": pygame.Rect(0, 0, 1, 1),
            "style": {},
            "id": 0,
            "children": [],
            "children_grid": [],
            "components": [],
            "parent": None,
            "top": False,
            "z": 0,
        }
        self._stack: dict[str, typing.Any] = self._parent
        self._parents_stack = [self._stack]
        self._id = 1
        self._memory: dict[int, dict[str, typing.Any]] = {}
        self._old_data: dict[int, dict[str, typing.Any]] = {}
        self._element: dict[str, typing.Any] = self._parent
        self._canva: pygame.Surface = None
        self._canva_rect: pygame.Rect = None
        self._abs_hovered: list[dict[str, typing.Any]] = []
        self._started = False
        self._started_pressing_element = None
        self._started_pressing_button = -1
        self._z = 0
        self._mouse_pos = pygame.Vector2()
        self._mouse_rel = pygame.Vector2()
        self._default_styles = {
            "element": {},
            "rect": {},
            "circle": {},
            "polygon": {},
            "line": {},
            "text": {},
            "image": {},
        }

    def _get_element(
        self, rect: _typing.RectLike, arg_style
    ) -> tuple[dict[str, typing.Any], "_data.Interaction"]:
        if arg_style is None:
            arg_style = {}
        style = self._default_styles["element"].copy()
        style.update(arg_style)
        if rect is None:
            rect = (0, 0, 0, 0)
        rect = pygame.Rect(rect)
        parent = self._parent
        if "parent_id" in style:
            if style["parent_id"] in self._memory:
                parent = self._memory[style["parent_id"]]
            elif style["parent_id"] == 0:
                parent = self._stack
        element: dict[str, typing.Any] = {
            "rect": rect,
            "abs_rect": rect.copy(),
            "style": style,
            "id": self._id,
            "children": [],
            "components": [],
            "children_grid": [],
            "parent": parent,
            "top": False,
            "z": 1,
            "hovered": False,
        }
        interaction = self._get_interaction(element)
        self._memory[self._id] = element
        if parent:
            z = self._z
            if "z" in style:
                z = style["z"]
            if parent["z"] > z:
                z = parent["z"] + len(parent["children"]) + 1
            parent["children"].append(element)
            if not self._style_val(style, "element", "ignore_grid", False):
                parent["children_grid"].append(element)
            element["z"] = z + 1
        self._id += 1
        self._element = element
        self._z += 1
        return element, interaction

    def _style_val(self, style: dict, major, minor, default):
        return style.get(minor, self._default_styles[major].get(minor, default))

    def _organize_grid(
        self,
        element,
        a,
        oa,
        av,
        oav,
        rect,
        style,
        pada,
        padoa,
        space,
        anchor,
        resizeoa,
        resizea,
        minresizeoa,
        minresizea,
        maxresizea,
        maxresizeoa,
        padded_oa,
        padded_a,
        children,
    ):
        grid_align = self._style_val(style, "element", "grid_align", "first")
        self._check_align(grid_align)
        spacea = _globalctx._abs_perc(
            self._style_val(style, "element", f"grid_space{a}", space),
            getattr(rect, av),
        )
        spaceoa = _globalctx._abs_perc(
            self._style_val(style, "element", f"grid_space{oa}", space),
            getattr(rect, oav),
        )

        line_elements = []
        lines = []
        line_size_a = 0
        longest_line_size_a = 0
        current_pos_oa = padoa
        longest_line_size_oa = 0
        for child in children:
            ch_rect = child["rect"]
            add = spacea
            if line_size_a <= 0:
                add = 0
            hypot_length_a = line_size_a + add + getattr(ch_rect, av)

            if hypot_length_a <= padded_a:
                line_elements.append(child)
                line_size_a = hypot_length_a
                if line_size_a > longest_line_size_a:
                    longest_line_size_a = line_size_a
                if getattr(ch_rect, oav) > longest_line_size_oa:
                    longest_line_size_oa = getattr(ch_rect, oav)
            else:
                if (
                    resizea
                    and hypot_length_a + pada * 2 <= maxresizea
                    and hypot_length_a + pada * 2 >= minresizea
                    and hypot_length_a > longest_line_size_a
                ):
                    longest_line_size_a = hypot_length_a
                    setattr(rect, av, hypot_length_a + pada * 2)
                    padded_a = hypot_length_a
                    line_elements.append(child)
                    line_size_a = hypot_length_a
                    if getattr(ch_rect, oav) > longest_line_size_oa:
                        longest_line_size_oa = getattr(ch_rect, oav)
                else:
                    lines.append([line_elements, longest_line_size_oa, line_size_a])
                    line_elements = []

                    line_elements.append(child)
                    longest_line_size_oa = getattr(ch_rect, oav)
                    line_size_a = getattr(ch_rect, av)
                    if current_pos_oa != padoa:
                        current_pos_oa += spaceoa
                    current_pos_oa += longest_line_size_oa

                    if (
                        resizeoa
                        and current_pos_oa + padoa * 2 <= maxresizeoa
                        and current_pos_oa + padoa * 2 >= minresizeoa
                    ):
                        setattr(rect, oav, current_pos_oa + padoa * 2)
                        padded_oa = current_pos_oa

        if len(line_elements) > 0:
            lines.append([line_elements, longest_line_size_oa, line_size_a])
            if current_pos_oa != padoa:
                current_pos_oa += spaceoa
            current_pos_oa += longest_line_size_oa
            if (
                resizeoa
                and current_pos_oa + padoa * 2 <= maxresizeoa
                and current_pos_oa + padoa * 2 >= minresizeoa
            ):
                setattr(rect, oav, current_pos_oa + padoa * 2)
                padded_oa = current_pos_oa

        lines_size_oa = current_pos_oa
        current_pos_oa = padoa
        spacing = spaceoa
        if anchor == "max_spacing":
            if len(lines) > 1:
                spacing = max(
                    0,
                    ((padded_oa - sum([line[1] for line in lines])) / (len(lines) - 1)),
                )
            else:
                anchor = "center"
        if anchor == "center":
            current_pos_oa = getattr(rect, oav) / 2 - lines_size_oa / 2
        if anchor == "last":
            lines = list(reversed(lines))
            current_pos_oa = padded_oa - lines_size_oa

        for elements, biggest_size, line_size in lines:
            current_pos_a = pada
            spacinga = spacea
            if grid_align == "max_spacing":
                if len(elements) > 1:
                    spacinga = max(
                        0,
                        (padded_a - sum([getattr(e["rect"], av) for e in elements]))
                        / (len(elements) - 1),
                    )
                else:
                    grid_align = "center"
            if grid_align == "last":
                current_pos_a = padded_a - line_size
            if grid_align == "center":
                current_pos_a = getattr(rect, av) / 2 - line_size / 2
            for el in elements:
                el_rect = el["rect"]
                setattr(el_rect, oa, current_pos_oa)
                setattr(el_rect, a, current_pos_a)
                current_pos_a += getattr(el_rect, av)
                current_pos_a += spacinga
            current_pos_oa += biggest_size
            current_pos_oa += spacing

        element["grid"] = {
            f"overflow{oa}": max(0, lines_size_oa - padded_oa),
            f"overflow{a}": max(0, longest_line_size_a - padded_a),
            f"pad{a}": pada,
            f"pad{oa}": padoa,
            "spacing": space,
        }

    def _check_align(self, align):
        if align not in ["first", "last", "center", "max_spacing"]:
            raise error.MILIValueError(f"Invalid alignment/anchoring '{align}'")

    def _organize_element(self, element: dict[str, typing.Any]):
        style = element["style"]
        rect = element["rect"]
        children = element["children_grid"]
        a = self._style_val(style, "element", "axis", "y")
        if a not in ["x", "y"]:
            raise error.MILIValueError(f"Invalid axis {a}")
        oa = "x" if a == "y" else "y"
        av = "w" if a == "x" else "h"
        oav = "w" if av == "h" else "h"
        pada = self._style_val(style, "element", f"pad{a}", 5)
        padoa = self._style_val(style, "element", f"pad{oa}", pada)
        pada, padoa = (
            _globalctx._abs_perc(pada, getattr(rect, av)),
            _globalctx._abs_perc(padoa, getattr(rect, oav)),
        )
        space = _globalctx._abs_perc(
            self._style_val(style, "element", "spacing", 3), getattr(rect, av)
        )
        anchor = self._style_val(style, "element", "anchor", "first")
        self._check_align(anchor)
        resizeoa = self._style_val(style, "element", f"resize{oa}", False)
        resizea = self._style_val(style, "element", f"resize{a}", False)
        if isinstance(resizea, dict):
            maxv = resizea.get("max", float("inf"))
            minv = resizea.get("min", 0)
            minresizea = _globalctx._abs_perc(minv, getattr(rect, av))
            maxresizea = _globalctx._abs_perc(maxv, getattr(rect, av))
            resizea = True
        else:
            resizea = bool(resizea)
            minresizea = 0
            maxresizea = float("inf")
        if isinstance(resizeoa, dict):
            maxv = resizeoa.get("max", float("inf"))
            minv = resizeoa.get("min", 0)
            minresizeoa = _globalctx._abs_perc(minv, getattr(rect, oav))
            maxresizeoa = _globalctx._abs_perc(maxv, getattr(rect, oav))
            resizeoa = True
        else:
            resizeoa = bool(resizeoa)
            minresizeoa = 0
            maxresizeoa = float("inf")
        _filloa, _filla = (
            self._style_val(style, "element", f"fill{oa}", False),
            self._style_val(style, "element", f"fill{a}", False),
        )
        if resizeoa and bool(_filloa) is not False:
            raise error.MILIIncompatibleStylesError(
                f"Cannot have resize{oa} True and fill{oa} not False"
            )
        if resizea and bool(_filla) is not False:
            raise error.MILIIncompatibleStylesError(
                f"Cannot have resize{a} True and fill{a} not False"
            )

        padded_oa = getattr(rect, oav) - padoa * 2
        padded_a = getattr(rect, av) - pada * 2

        grid = self._style_val(style, "element", "grid", False)
        if grid:
            self._organize_grid(
                element,
                a,
                oa,
                av,
                oav,
                rect,
                style,
                pada,
                padoa,
                space,
                anchor,
                resizeoa,
                resizea,
                minresizeoa,
                minresizea,
                maxresizea,
                maxresizeoa,
                padded_oa,
                padded_a,
                children,
            )
            return

        elements_with_filloa = []
        elements_with_filla = []
        biggest_oa = 0
        fixed_elements_a = 0

        for el in children:
            el_rect = el["rect"]
            el_style = el["style"]
            el_filloa = self._style_val(el_style, "element", f"fill{oa}", False)
            el_filla = self._style_val(el_style, "element", f"fill{a}", False)
            if bool(el_filloa) is False and getattr(el_rect, oav) > biggest_oa:
                biggest_oa = getattr(el_rect, oav)
            if bool(el_filloa) is True:
                elements_with_filloa.append(el)
            fixed_elements_a += space
            if bool(el_filla) is False:
                fixed_elements_a += getattr(el_rect, av)
            else:
                elements_with_filla.append(el)
        fixed_elements_a -= space
        available_to_filla = max(0, padded_a - fixed_elements_a)

        if resizeoa and biggest_oa > 0:
            setattr(
                rect, oav, min(max(biggest_oa + padoa * 2, minresizeoa), maxresizeoa)
            )
            padded_oa = getattr(rect, oav) - padoa * 2
        if resizea and fixed_elements_a > 0:
            setattr(
                rect, av, min(max(fixed_elements_a + pada * 2, minresizea), maxresizea)
            )
            padded_a = getattr(rect, av) - pada * 2
            available_to_filla = 0

        changed = []

        for filloa_el in elements_with_filloa:
            el_rect = filloa_el["rect"]
            el_style = filloa_el["style"]
            filloa = self._style_val(el_style, "element", f"fill{oa}", False)
            if filloa is True:
                filloa = "100"
            el_filloa = _globalctx._abs_perc(filloa, padded_oa)
            setattr(el_rect, oav, el_filloa)
            changed.append(filloa_el)

        filla_totalsize = 0
        for filla_el in elements_with_filla:
            el_rect = filla_el["rect"]
            el_style = filla_el["style"]
            filla = self._style_val(el_style, "element", f"fill{a}", False)
            if filla is True:
                filla = "100"
            el_filla = _globalctx._abs_perc(
                filla,
                available_to_filla,
            )
            filla_el["filla"] = el_filla
            filla_totalsize += el_filla

        total_a = fixed_elements_a
        if filla_totalsize != 0:
            if fixed_elements_a <= padded_a:
                total_a = padded_a
            else:
                total_a = fixed_elements_a
            multiplier = 1
            if filla_totalsize > available_to_filla:
                multiplier = available_to_filla / filla_totalsize
            else:
                total_a = fixed_elements_a + filla_totalsize
            for filla_el in elements_with_filla:
                el_rect = filla_el["rect"]
                setattr(el_rect, av, filla_el["filla"] * multiplier)
                changed.append(filla_el)
                del filla_el["filla"]

        current_a = pada
        spacing = space
        if anchor == "max_spacing":
            if len(children) > 1:
                spacing = max(
                    0,
                    (
                        (padded_a - sum([getattr(e["rect"], av) for e in children]))
                        / (len(children) - 1)
                    ),
                )
            else:
                anchor = "center"
        if anchor == "last":
            current_a = getattr(rect, av) - pada - total_a
        if anchor == "center":
            current_a = getattr(rect, av) / 2 - total_a / 2
        for el in children:
            el_rect = el["rect"]
            el_style = el["style"]
            el_align = self._style_val(el_style, "element", "align", "first")
            self._check_align(el_align)
            setattr(el_rect, a, current_a)
            current_a += getattr(el_rect, av)
            current_a += spacing
            if el_align == "first":
                setattr(el_rect, oa, padoa)
            if el_align == "last":
                setattr(
                    el_rect, "right" if a == "y" else "left", getattr(rect, oav) - padoa
                )
            if el_align == "center":
                setattr(el_rect, f"center{oa}", getattr(rect, oav) / 2)

        element["grid"] = {
            f"overflow{oa}": max(0, biggest_oa - padded_oa),
            f"overflow{a}": max(0, total_a - padded_a),
            f"pad{a}": pada,
            f"pad{oa}": padoa,
            "spacing": space,
        }

        for el in list(changed):
            self._organize_element(el)

    def _start(self, style):
        self._parent = {
            "rect": pygame.Rect((0, 0), self._canva.size),
            "style": style,
            "id": 0,
            "children": [],
            "children_grid": [],
            "components": [],
            "parent": None,
            "top": False,
            "z": 0,
        }
        self._id = 1
        self._element = self._parent
        self._stack = self._parent
        self._parents_stack = [self._stack]
        self._abs_hovered = []
        self._started = True
        self._z = 0

        if (btn := _globalctx._get_first_button(pygame.mouse.get_just_pressed())) > -1:
            self._started_pressing_button = btn
        if (
            _globalctx._get_first_button(pygame.mouse.get_just_released())
            == self._started_pressing_button
        ):
            self._started_pressing_button = -1
            self._started_pressing_element = None

    def _check_interaction(self, element: dict[str, typing.Any]):
        if not self._style_val(element["style"], "element", "blocking", True):
            return
        clipdraw = self._style_val(element["style"], "element", "clip_draw", True)
        hover = element["abs_rect"].collidepoint(pygame.mouse.get_pos())
        parent_hover = True
        if element["parent"] is not None:
            parent_hover = element["parent"]["abs_rect"].collidepoint(
                pygame.mouse.get_pos()
            )
            if clipdraw:
                parent_hover = True
        if hover and parent_hover:
            self._abs_hovered.append(element)

    def _get_interaction(self, element: dict[str, typing.Any]) -> "_data.Interaction":
        if element["id"] not in self._memory:
            return _data.Interaction(
                self._mili, -1, False, -1, -1, -1, False, False, False, False
            )
        old_el = self._memory[element["id"]]
        self._old_data[element["id"]] = {
            "rect": old_el["rect"],
            "abs_rect": old_el["abs_rect"],
            "components": old_el["components"],
            "children_ids": [ch["id"] for ch in old_el["children"]],
            "parent_id": old_el["parent"]["id"] if old_el["parent"] else 0,
            "grid": old_el["grid"] if "grid" in old_el else None,
        }
        absolute_hover = old_el["abs_rect"].collidepoint(pygame.mouse.get_pos())
        if old_el["top"]:
            if (
                self._started_pressing_button > -1
                and self._started_pressing_element is None
            ):
                self._started_pressing_element = element["id"]
                return self._total_interaction(
                    element, absolute_hover, absolute_hover, False, old_el
                )
            if (
                self._started_pressing_button > -1
                and element["id"] != self._started_pressing_element
            ):
                return self._partial_interaction(element, absolute_hover)
            return self._total_interaction(
                element, absolute_hover, absolute_hover, False, old_el
            )
        else:
            if (
                self._started_pressing_button > -1
                and element["id"] == self._started_pressing_element
            ):
                return self._total_interaction(
                    element, absolute_hover, False, True, old_el
                )
            return self._partial_interaction(element, absolute_hover)

    def _partial_interaction(self, element, absolute_hover):
        element["hovered"] = False
        return _data.Interaction(
            self._mili,
            element["id"],
            False,
            -1,
            -1,
            -1,
            absolute_hover,
            False,
            False,
            False,
        )

    def _total_interaction(self, element, absolute_hover, hovered, unhovered, old_el):
        i = _data.Interaction(
            self._mili,
            element["id"],
            absolute_hover and hovered,
            _globalctx._get_first_button(pygame.mouse.get_pressed(5)),
            _globalctx._get_first_button(pygame.mouse.get_just_pressed()),
            _globalctx._get_first_button(pygame.mouse.get_just_released()),
            absolute_hover,
            unhovered,
            False,
            False,
        )
        if i.hovered and not old_el["hovered"]:
            i.just_hovered = True
        if not i.hovered and old_el["hovered"]:
            i.just_unhovered = True
        element["hovered"] = i.hovered
        return i

    def _add_component(self, type, data, arg_style):
        self._start_check()
        if arg_style is None:
            arg_style = {}
        style = self._default_styles[type].copy()
        style.update(arg_style)
        if not self._element:
            raise error.MILIStatusError(
                "Cannot add component if no element was created"
            )
        if type == "text":
            self._text_resize(self._element["rect"], data, style)
        if _globalctx._component_types[type] != "builtin":
            _globalctx._component_types[type].added(self, data, style, self._element)
        self._element["components"].append({"type": type, "data": data, "style": style})

    def _get_font(self, style) -> pygame.Font:
        path = self._style_val(style, "text", "name", None)
        size = self._style_val(style, "text", "size", 20)
        key = f"{path}_{size}"
        if key not in _globalctx._font_cache:
            sysfont = self._style_val(style, "text", "sysfont", False)
            func = pygame.font.SysFont if sysfont else pygame.font.Font
            font = func(path, size)
            _globalctx._font_cache[key] = font
        return _globalctx._font_cache[key]

    def _draw_comp_rect(self, data, style, el, rect: pygame.Rect):
        padx = self._style_val(style, "rect", "padx", 0)
        pady = self._style_val(style, "rect", "pady", padx)
        padx, pady = (
            _globalctx._abs_perc(padx, rect.w),
            _globalctx._abs_perc(pady, rect.h),
        )
        outline = _globalctx._abs_perc(
            self._style_val(style, "rect", "outline", 0), min(rect.w, rect.h)
        )
        border_radius = _globalctx._abs_perc(
            self._style_val(style, "rect", "border_radius", 7), min(rect.w, rect.h)
        )
        color = self._style_val(style, "rect", "color", "black")
        pygame.draw.rect(
            self._canva,
            color,
            rect.inflate(-padx, -pady),
            int(outline),
            int(border_radius),
        )

    def _draw_comp_circle(self, data, style, el, rect: pygame.Rect):
        padx = self._style_val(style, "circle", "padx", 0)
        pady = self._style_val(style, "circle", "pady", padx)
        padx, pady = (
            _globalctx._abs_perc(padx, rect.w),
            _globalctx._abs_perc(pady, rect.h),
        )
        outline = _globalctx._abs_perc(
            self._style_val(style, "circle", "outline", 0), min(rect.w, rect.h)
        )
        color = self._style_val(style, "circle", "color", "black")
        antialias = self._style_val(style, "circle", "antialias", False)
        if padx == pady:
            (pygame.draw.aacircle if antialias else pygame.draw.circle)(
                self._canva, color, rect.center, rect.w / 2 - padx, int(outline)
            )
        else:
            pygame.draw.ellipse(
                self._canva, color, rect.inflate(-padx, -pady), int(outline)
            )

    def _draw_comp_polygon(
        self, data: list[tuple[int, int]], style, el, rect: pygame.Rect
    ):
        points = []
        for raw_p in data:
            rx, ry = raw_p
            rx, ry = _globalctx._abs_perc(rx, rect.w), _globalctx._abs_perc(ry, rect.h)
            points.append((rect.centerx + rx, rect.centery + ry))
        color = self._style_val(style, "polygon", "color", "black")
        outline = _globalctx._abs_perc(
            self._style_val(style, "polygon", "outline", 0), min(rect.w, rect.h)
        )
        pygame.draw.polygon(self._canva, color, points, int(outline))

    def _draw_comp_line(
        self, data: list[tuple[int, int]], style, el, rect: pygame.Rect
    ):
        points = []
        for raw_p in data:
            rx, ry = raw_p
            rx, ry = _globalctx._abs_perc(rx, rect.w), _globalctx._abs_perc(ry, rect.h)
            points.append((rect.centerx + int(rx), rect.centery + int(ry)))
        if len(points) != 2:
            raise error.MILIValueError("Wrong number of points")
        color = self._style_val(style, "line", "color", "black")
        size = _globalctx._abs_perc(
            self._style_val(style, "line", "size", 1), min(rect.w, rect.h)
        )
        antialias = self._style_val(style, "line", "antialias", False)
        (pygame.draw.aaline if antialias else pygame.draw.line)(
            self._canva, color, points[0], points[1], max(1, int(size))
        )

    def _draw_comp_text(self, data: str, style, el, rect: pygame.Rect):
        font = self._get_font(style)
        font.align = self._style_val(style, "text", "font_align", pygame.FONT_CENTER)
        font.bold = self._style_val(style, "text", "bold", False)
        font.italic = self._style_val(style, "text", "italic", False)
        font.underline = self._style_val(style, "text", "underline", False)
        font.strikethrough = self._style_val(style, "text", "strikethrough", False)
        antialias = self._style_val(style, "text", "antialias", True)
        color = self._style_val(style, "text", "color", "white")
        bg_color = self._style_val(style, "text", "bg_color", None)
        align = self._style_val(style, "text", "align", "center")
        growx, growy = (
            self._style_val(style, "text", "growx", False),
            self._style_val(style, "text", "growy", True),
        )
        padx = self._style_val(style, "text", "padx", 5)
        pady = self._style_val(style, "text", "pady", 3)
        padx, pady = (
            _globalctx._abs_perc(padx, rect.w),
            _globalctx._abs_perc(pady, rect.h),
        )
        wraplen = _globalctx._abs_perc(
            self._style_val(style, "text", "wraplen", 0), rect.w - padx * 2
        )

        surf = font.render(str(data), antialias, color, bg_color, max(0, int(wraplen)))
        sw, sh = surf.size
        if growy and sh > rect.h + pady * 2:
            rect.h = sh + pady * 2
        if growx and sw > rect.w + padx * 2:
            rect.w = sw + padx * 2

        txtrect = surf.get_rect(**_globalctx._text_align(align, rect, padx, pady))
        self._canva.blit(surf, txtrect)

    def _text_size(self, data: str, style):
        font = self._get_font(style)
        font.align = self._style_val(style, "text", "font_align", pygame.FONT_CENTER)
        font.bold = self._style_val(style, "text", "bold", False)
        font.italic = self._style_val(style, "text", "italic", False)
        font.underline = self._style_val(style, "text", "underline", False)
        font.strikethrough = self._style_val(style, "text", "strikethrough", False)
        if self._style_val(style, "text", "slow_grow", False):
            wraplen = _globalctx._abs_perc(
                self._style_val(style, "text", "wraplen", 0), 0
            )
            return font.render(
                data, True, "white", None, max(0, int(wraplen))
            ).get_size()
        return font.size(data)

    def _text_resize(self, rect: pygame.Rect, data, style):
        growx, growy = (
            self._style_val(style, "text", "growx", False),
            self._style_val(style, "text", "growy", True),
        )
        if not growx and not growy:
            return
        slow_grow = self._style_val(style, "text", "slow_grow", False)
        padx = self._style_val(style, "text", "padx", 5)
        pady = self._style_val(style, "text", "pady", 3)
        padx, pady = (
            _globalctx._abs_perc(padx, rect.w),
            _globalctx._abs_perc(pady, rect.h),
        )
        font = self._get_font(style)
        if slow_grow and not growx:
            font.align = self._style_val(
                style, "text", "font_align", pygame.FONT_CENTER
            )
            font.bold = self._style_val(style, "text", "bold", False)
            font.italic = self._style_val(style, "text", "italic", False)
            font.underline = self._style_val(style, "text", "underline", False)
            font.strikethrough = self._style_val(style, "text", "strikethrough", False)
            wraplen = _globalctx._abs_perc(
                self._style_val(style, "text", "wraplen", 0), rect.w - padx * 2
            )
            sw, sh = font.render(
                data, True, "white", None, max(0, int(wraplen))
            ).get_size()
        else:
            sw, sh = font.size(str(data))
        if growy and sh > rect.h + pady * 2:
            rect.h = sh + pady * 2
        if growx and sw > rect.w + padx * 2:
            rect.w = sw + padx * 2

    def _draw_comp_image(self, data: pygame.Surface, style, el, rect: pygame.Rect):
        if data is None:
            return
        cache = style.get("cache", None)
        padx = self._style_val(style, "image", "padx", 0)
        pady = self._style_val(style, "image", "pady", padx)
        padx, pady = (
            _globalctx._abs_perc(padx, rect.w),
            _globalctx._abs_perc(pady, rect.h),
        )
        do_fill = self._style_val(style, "image", "fill", False)
        do_stretchx = self._style_val(style, "image", "stretchx", False)
        do_stretchy = self._style_val(style, "image", "stretchy", False)
        fill_color = self._style_val(style, "image", "fill_color", None)
        smoothscale = self._style_val(style, "image", "smoothscale", False)
        border_radius = _globalctx._abs_perc(
            self._style_val(style, "image", "border_radius", 0), min(rect.w, rect.h)
        )
        alpha = self._style_val(style, "image", "alpha", 255)
        nine_patch = _globalctx._abs_perc(
            self._style_val(style, "image", "ninepatch_size", 0), min(rect.w, rect.h)
        )
        nine_patch = pygame.math.clamp(nine_patch, 0, min(rect.w, rect.h) / 2)
        if nine_patch != 0 and (do_fill or do_stretchx or do_stretchy):
            raise error.MILIIncompatibleStylesError(
                "9-patch image mode is incompatible with fill, stretchx, and stretchy styles"
            )

        if not cache:
            output = _globalctx._get_image(
                rect,
                data,
                padx,
                pady,
                do_fill,
                do_stretchx,
                do_stretchy,
                fill_color,
                border_radius,
                alpha,
                smoothscale,
                nine_patch,
            )
        else:
            new_cache = {
                "data": data,
                "padx": padx,
                "pady": pady,
                "fill": do_fill,
                "stretchx": do_stretchx,
                "stretchy": do_stretchy,
                "fill_color": fill_color,
                "border_radius": border_radius,
                "alpha": alpha,
                "size": rect.size,
                "smoothscale": smoothscale,
                "nine_patch": nine_patch,
            }
            if cache._cache is None:
                output = _globalctx._get_image(
                    rect,
                    data,
                    padx,
                    pady,
                    do_fill,
                    do_stretchx,
                    do_stretchy,
                    fill_color,
                    border_radius,
                    alpha,
                    smoothscale,
                    nine_patch,
                )
                new_cache["output"] = output
                cache._cache = new_cache
            else:
                _cache = cache._cache
                if (
                    data is not _cache["data"]
                    or smoothscale != _cache["smoothscale"]
                    or rect.size != _cache["size"]
                    or padx != _cache["padx"]
                    or pady != _cache["pady"]
                    or do_fill != _cache["fill"]
                    or do_stretchx != _cache["stretchx"]
                    or do_stretchy != _cache["stretchy"]
                    or fill_color != _cache["fill_color"]
                    or border_radius != _cache["border_radius"]
                    or alpha != _cache["alpha"]
                    or nine_patch != _cache["nine_patch"]
                ):
                    output = _globalctx._get_image(
                        rect,
                        data,
                        padx,
                        pady,
                        do_fill,
                        do_stretchx,
                        do_stretchy,
                        fill_color,
                        border_radius,
                        alpha,
                        smoothscale,
                        nine_patch,
                    )
                    new_cache["output"] = output
                    cache._cache = new_cache
                else:
                    output = _cache["output"]

        image_rect = output.get_rect(center=rect.center)
        self._canva.blit(output, image_rect)

    def _draw_update_element(
        self, element: dict[str, typing.Any], parent_pos, parent_clip=None
    ):
        el_style = element["style"]
        offset = self._style_val(el_style, "element", "offset", (0, 0))
        do_clip = self._style_val(el_style, "element", "clip_draw", True)
        absolute_rect = element["rect"].move(
            (parent_pos[0] + offset[0], parent_pos[1] + offset[1])
        )
        if absolute_rect.w <= 0 or absolute_rect.h <= 0:
            return
        if parent_clip is None:
            parent_clip = absolute_rect
        if not absolute_rect.colliderect(parent_clip):
            return
        if do_clip:
            clip = absolute_rect.clip(parent_clip)
        else:
            clip = parent_clip
        element["abs_rect"] = absolute_rect
        self._check_interaction(element)

        pre_draw, mid_draw, post_draw = [
            self._style_val(el_style, "element", name, None)
            for name in ["pre_draw_func", "mid_draw_func", "post_draw_func"]
        ]
        el_data = None

        self._canva.set_clip(clip)
        if pre_draw:
            if el_data is None:
                el_data = _globalctx._element_data(self._mili, element, None)
            pre_draw(self._canva, el_data, clip)
            self._canva.set_clip(clip)
        render_above = []
        for component in element["components"]:
            comp_type = component["type"]
            if self._style_val(component["style"], comp_type, "draw_above", False):
                render_above.append(component)
                continue
            if _globalctx._component_types[comp_type] == "builtin":
                getattr(self, f"_draw_comp_{comp_type}")(
                    component["data"], component["style"], element, absolute_rect
                )
            else:
                _globalctx._component_types[comp_type].draw(
                    self, component["data"], component["style"], element, absolute_rect
                )
        if mid_draw:
            if el_data is None:
                el_data = _globalctx._element_data(self._mili, element, None)
            mid_draw(self._canva, el_data, clip)
            self._canva.set_clip(clip)
        if len(element["children"]) > 0:
            for child in sorted(element["children"], key=lambda c: c["z"]):
                self._canva.set_clip(clip)
                self._draw_update_element(child, absolute_rect.topleft, clip)
                self._canva.set_clip(clip)
        for component in render_above:
            comp_type = component["type"]
            if _globalctx._component_types[comp_type] == "builtin":
                getattr(self, f"_draw_comp_{comp_type}")(
                    component["data"], component["style"], element, absolute_rect
                )
            else:
                _globalctx._component_types[comp_type].draw(
                    self, component["data"], component["style"], element, absolute_rect
                )
        if post_draw:
            if el_data is None:
                el_data = _globalctx._element_data(self._mili, element, None)
            post_draw(self._canva, el_data, clip)
            self._canva.set_clip(clip)

    def _start_check(self):
        if not self._started:
            raise error.MILIStatusError("MILI.start() was not called")



import pygame
import typing

from mili import _core
from mili import error as _error
from mili import data as _data
from mili import typing as _typing


__all__ = ("MILI", "register_custom_component", "pack_component")


def register_custom_component(name: str, component_type: _typing.ComponentProtocol):
    _core._globalctx._component_types[name] = component_type


def pack_component(name: str, data: typing.Any, style: _typing.AnyStyleLike):
    return {"type": name, "data": data, "style": style}


class MILI:
    def __init__(self, canva: pygame.Surface | None = None):
        self._ctx = _core._ctx(self)
        self.always_get_data: bool = False
        if canva is not None:
            self.set_canva(canva)

    @property
    def stack_id(self) -> int:
        return self._ctx._stack["id"]

    @property
    def current_parent_id(self) -> int:
        return self._ctx._parent["id"] if self._ctx._parent else -1

    @property
    def all_elements_ids(self) -> list[int]:
        return list(self._ctx._memory.keys())

    @property
    def canva(self) -> pygame.Surface:
        return self._ctx._canva

    @canva.setter
    def canva(self, v: pygame.Surface):
        self.set_canva(v)

    def default_style(self, type: str, style: _typing.AnyStyleLike):
        if type not in _core._globalctx._component_types:
            raise _error.MILIValueError("Invalid style type")
        else:
            if type not in self._ctx._default_styles:
                self._ctx._default_styles[type] = {}
        self._ctx._default_styles[type].update(style)

    def default_styles(self, **types_styles: _typing.AnyStyleLike):
        for name, value in types_styles.items():
            self.default_style(name, value)

    def reset_style(self, *types: str):
        for tp in types:
            if tp not in _core._globalctx._component_types:
                raise _error.MILIValueError("Invalid style type")
            self._ctx._default_styles[tp] = {}

    def start(
        self, style: _typing.ElementStyleLike | None = None, is_global=True
    ) -> typing.Literal[True]:
        if self._ctx._canva is None:
            raise _error.MILIStatusError("Canva was not set")
        _core._globalctx._mili_stack.append(self)
        _core._globalctx._mili = self
        if style is None:
            style = {}
        style.update(blocking=False)
        self._ctx._start(style)
        mouse_pos = pygame.Vector2(pygame.mouse.get_pos())
        self._ctx._mouse_rel = mouse_pos - self._ctx._mouse_pos
        self._ctx._mouse_pos = mouse_pos
        if is_global:
            _data.ImageCache._preallocated_index = -1
        return True

    def set_canva(self, surface: pygame.Surface):
        self._ctx._canva = surface
        self._ctx._canva_rect = surface.get_rect()

    def update_draw(self):
        self._ctx._organize_element(self._ctx._stack)
        self._ctx._started = False
        self._ctx._draw_update_element(self._ctx._stack, (0, 0))
        abs_hovered = sorted(self._ctx._abs_hovered, key=lambda e: e["z"], reverse=True)
        if len(abs_hovered) > 0:
            abs_hovered[0]["top"] = True
        if len(_core._globalctx._mili_stack) > 0:
            _core._globalctx._mili_stack.pop()
        if len(_core._globalctx._mili_stack) > 0:
            _core._globalctx._mili = _core._globalctx._mili_stack[-1]
        else:
            _core._globalctx._mili = None

    def element(
        self,
        rect: _typing.RectLike | None,
        style: _typing.ElementStyleLike | None = None,
        get_data: bool = False,
    ) -> _data.Interaction | _data.ElementData:
        self._ctx._start_check()
        el, interaction = self._ctx._get_element(rect, style)
        if get_data or self.always_get_data:
            return _core._globalctx._element_data(self, el, interaction)
        return interaction

    def begin(
        self,
        rect: _typing.RectLike | None,
        style: _typing.ElementStyleLike | None = None,
        header: str = "",
        get_data: bool = False,
    ) -> _data.Interaction | _data.ElementData:
        self._ctx._start_check()
        el, interaction = self._ctx._get_element(rect, style)
        self._ctx._parent = el
        self._ctx._parents_stack.append(el)
        if get_data or self.always_get_data:
            return _core._globalctx._element_data(self, el, interaction)
        return interaction

    def end(self, header: str = ""):
        self._ctx._start_check()
        if self._ctx._parent and self._ctx._parent["id"] != 0:
            self._ctx._organize_element(self._ctx._parent)
        else:
            raise _error.MILIStatusError("end() called too many times")
        if len(self._ctx._parents_stack) > 1:
            self._ctx._parents_stack.pop()
            self._ctx._parent = self._ctx._parents_stack[-1]
        else:
            self._ctx._parent = self._ctx._parents_stack[0]

    def rect(self, style: _typing.RectStyleLike | None = None):
        self._ctx._add_component("rect", None, style)

    def rect_element(
        self,
        rect_style: _typing.RectStyleLike | None = None,
        element_rect: _typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
        get_data: bool = False,
    ) -> _data.Interaction | _data.ElementData:
        data = self.element(element_rect, element_style, get_data)
        self.rect(rect_style)
        return data

    def circle(self, style: _typing.CircleStyleLike | None = None):
        self._ctx._add_component("circle", None, style)

    def circle_element(
        self,
        circle_style: _typing.CircleStyleLike | None = None,
        element_rect: _typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
        get_data: bool = False,
    ) -> _data.Interaction | _data.ElementData:
        data = self.element(element_rect, element_style, get_data)
        self.circle(circle_style)
        return data

    def polygon(
        self, points: _typing.PointsLike, style: _typing.PolygonStyleLike | None = None
    ):
        self._ctx._add_component("polygon", points, style)

    def polygon_element(
        self,
        points: _typing.PointsLike,
        polygon_style: _typing.PolygonStyleLike | None = None,
        element_rect: _typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
        get_data: bool = False,
    ) -> _data.Interaction | _data.ElementData:
        data = self.element(element_rect, element_style, get_data)
        self.polygon(points, polygon_style)
        return data

    def line(
        self,
        start_end: _typing.PointsLike,
        style: _typing.LineStyleLike | None = None,
    ):
        self._ctx._add_component("line", start_end, style)

    def line_element(
        self,
        start_end: _typing.PointsLike,
        line_style: _typing.LineStyleLike | None = None,
        element_rect: _typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
        get_data: bool = False,
    ) -> _data.Interaction | _data.ElementData:
        data = self.element(element_rect, element_style, get_data)
        self.line(start_end, line_style)
        return data

    def text(self, text: str, style: _typing.TextStyleLike | None = None):
        self._ctx._add_component("text", str(text), style)

    def text_element(
        self,
        text: str,
        text_style: _typing.TextStyleLike | None = None,
        element_rect: _typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
        get_data: bool = False,
    ) -> _data.Interaction | _data.ElementData:
        data = self.element(element_rect, element_style, get_data)
        self.text(text, text_style)
        return data

    def text_size(
        self, text: str, style: _typing.TextStyleLike | None = None
    ) -> pygame.Vector2:
        if style is None:
            style = {}
        return pygame.Vector2(self._ctx._text_size(text, style))

    def text_font(self, style: _typing.TextStyleLike | None = None) -> pygame.Font:
        if style is None:
            style = {}
        return self._ctx._get_font(style)

    def image(
        self, surface: pygame.Surface, style: _typing.ImageStyleLike | None = None
    ):
        self._ctx._add_component("image", surface, style)

    def image_element(
        self,
        surface: pygame.Surface,
        image_style: _typing.ImageStyleLike | None = None,
        element_rect: _typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
        get_data: bool = False,
    ) -> _data.Interaction | _data.ElementData:
        data = self.element(element_rect, element_style, get_data)
        self.image(surface, image_style)
        return data

    def basic_element(
        self,
        rect: _typing.RectLike | None = None,
        style: _typing.ElementStyleLike | None = None,
        comp_data=None,
        comp_style: _typing.AnyStyleLike | None = None,
        bg_style: _typing.RectStyleLike | None = None,
        outline_style: _typing.RectStyleLike | None = None,
        component="text",
        get_data: bool = False,
    ) -> _data.Interaction | _data.ElementData:
        if component not in _core._globalctx._component_types.keys():
            raise _error.MILIValueError("Invalid component name")
        if not outline_style:
            outline_style = {}
        if "outline" not in outline_style:
            outline_style["outline"] = 1
        data = self.element(rect, style, get_data)
        self.rect(bg_style)
        comp = getattr(self, component)
        try:
            comp(comp_data, comp_style)
        except TypeError:
            comp(comp_style)
        self.rect(outline_style)
        return data

    def custom_component(
        self, component: str, data: typing.Any, style: dict[str, typing.Any] | None
    ):
        if (
            component not in self._ctx._default_styles
            and component in _core._globalctx._component_types
        ):
            self._ctx._default_styles[component] = {}
        elif component not in _core._globalctx._component_types:
            raise _error.MILIStatusError(
                f"No custom component '{component}' registered"
            )
        self._ctx._add_component(component, data, style)

    def packed_component(
        self,
        *packed_components: dict[typing.Literal["type", "style", "data"], typing.Any],
    ):
        for comp in packed_components:
            if comp["type"] not in _core._globalctx._component_types:
                _error.MILIStatusError(f"No component '{comp['type']}' exists")
            self._ctx._add_component(comp["type"], comp["data"], comp["style"])

    def data_from_id(self, element_id: int) -> _data.ElementData:
        self._ctx._start_check()
        if element_id in self._ctx._memory:
            el = self._ctx._memory[element_id]
            return _core._globalctx._element_data(el, self._ctx._get_interaction(el))

    def clear_memory(self):
        self._ctx._memory = {}

    def id_checkpoint(self, id_: int):
        if id_ < 1:
            id_ = 1
        if self._ctx._id < id_:
            self._ctx._id = id_



import typing
import pygame
from mili import data as _data

if typing.TYPE_CHECKING:
    from mili import _core

__all__ = (
    "NumberOrPercentage",
    "ColorLike",
    "RectLike",
    "PointsLike",
    "AnimValueLike",
    "EasingLike",
    "ElementStyleLike",
    "RectStyleLike",
    "CircleStyleLike",
    "LineStyleLike",
    "PolygonStyleLike",
    "TextStyleLike",
    "ImageStyleLike",
    "AnyStyleLike",
    "ComponentProtocol",
)

type NumberOrPercentage = int | float | str
type ColorLike = int | str | typing.Sequence[int] | pygame.Color
type RectLike = typing.Sequence[float | typing.Sequence[float]] | pygame.Rect
type PointsLike = typing.Sequence[typing.Sequence[NumberOrPercentage]]
type AnimValueLike = float | pygame.Color | typing.Sequence[float]
type EasingLike = typing.Callable[[float], float]


class ElementStyleLike(typing.TypedDict):
    resizex: bool | dict[typing.Literal["min", "max"], NumberOrPercentage]
    resizey: bool | dict[typing.Literal["min", "max"], NumberOrPercentage]
    fillx: bool | NumberOrPercentage
    filly: bool | NumberOrPercentage
    blocking: bool
    ignore_grid: bool
    align: typing.Literal["first", "center", "last"]
    z: int
    parent_id: int
    offset: typing.Sequence[float]
    clip_draw: bool
    pre_draw_func: (
        typing.Callable[[pygame.Surface, _data.ElementData, pygame.Rect], None] | None
    )
    mid_draw_func: (
        typing.Callable[[pygame.Surface, _data.ElementData, pygame.Rect], None] | None
    )
    post_draw_func: (
        typing.Callable[[pygame.Surface, _data.ElementData, pygame.Rect], None] | None
    )

    axis: typing.Literal["x", "y"]
    spacing: NumberOrPercentage
    padx: NumberOrPercentage
    pady: NumberOrPercentage
    anchor: typing.Literal["first", "center", "last", "max_spacing"]
    grid: bool
    grid_align: typing.Literal["first", "center", "last", "max_spacing"]
    grid_spacex: NumberOrPercentage
    grid_spacey: NumberOrPercentage


class RectStyleLike(typing.TypedDict):
    padx: NumberOrPercentage
    pady: NumberOrPercentage
    outline: NumberOrPercentage
    border_radius: NumberOrPercentage
    color: ColorLike
    draw_above: bool


class CircleStyleLike(typing.TypedDict):
    padx: NumberOrPercentage
    pady: NumberOrPercentage
    outline: NumberOrPercentage
    color: ColorLike
    antialias: bool
    draw_above: bool


class LineStyleLike(typing.TypedDict):
    size: NumberOrPercentage
    color: ColorLike
    antialias: bool
    draw_above: bool


class PolygonStyleLike(typing.TypedDict):
    outline: NumberOrPercentage
    color: ColorLike
    draw_above: bool


class TextStyleLike(typing.TypedDict):
    name: str | None
    size: int
    sysfont: bool
    align: typing.Literal[
        "center",
        "topleft",
        "bottomleft",
        "topright",
        "bottomright",
        "midleft",
        "left",
        "midright",
        "right",
        "midtop",
        "top",
        "midbottom",
        "bottom",
    ]
    font_align: int
    bold: bool
    italic: bool
    underline: bool
    strikethrough: bool
    antialias: bool
    color: ColorLike
    bg_color: ColorLike
    growx: bool
    growy: bool
    padx: NumberOrPercentage
    pady: NumberOrPercentage
    wraplen: NumberOrPercentage
    draw_above: bool
    slow_grow: bool


class ImageStyleLike(typing.TypedDict):
    cache: _data.ImageCache | None
    padx: NumberOrPercentage
    pady: NumberOrPercentage
    fill: bool
    stretchx: bool
    stretchy: bool
    fill_color: ColorLike
    smoothscale: bool
    border_radius: NumberOrPercentage
    alpha: int
    ninepatch_size: NumberOrPercentage
    draw_above: bool


type AnyStyleLike = (
    ElementStyleLike
    | RectStyleLike
    | CircleStyleLike
    | LineStyleLike
    | PolygonStyleLike
    | TextStyleLike
    | ImageStyleLike
    | dict[str]
)


class ComponentProtocol(typing.Protocol):
    def draw(
        self,
        ctx: "_core._ctx",
        data: typing.Any,
        style: dict[str],
        element: dict[str, typing.Any],
        absolute_rect: pygame.Rect,
    ): ...

    def added(
        self,
        ctx: "_core._ctx",
        data: typing.Any,
        style: dict[str],
        element: dict[str, typing.Any],
    ): ...



import pygame
import typing
from mili import _core
from mili import data as _data
from mili import typing as _typing
from mili import error as _error
from mili.mili import MILI as _MILI

__all__ = (
    "Selectable",
    "Dragger",
    "Scroll",
    "Scrollbar",
    "Slider",
    "GenericApp",
    "InteractionSound",
    "percentage",
    "indent",
    "fit_image",
)


class Selectable:
    def __init__(self, selected: bool = False):
        self.selected: bool = selected

    def update[IT: _data.Interaction | _data.ElementData](
        self,
        element: IT,
        selectable_group: typing.Sequence["Selectable"] | None = None,
        can_deselect_group: bool = True,
    ) -> IT:
        if selectable_group is None:
            selectable_group = []
        interaction = element
        if isinstance(element, _data.ElementData):
            interaction = element.interaction
        if interaction.left_just_released:
            self.selected = not self.selected
            if len(selectable_group) > 0:
                if self.selected:
                    for sel in selectable_group:
                        if sel is not self:
                            sel.selected = False
                elif not can_deselect_group:
                    one = False
                    for sel in selectable_group:
                        if sel is not self and sel.selected:
                            one = True
                            break
                    if not one:
                        self.selected = True
        elif len(selectable_group) > 0:
            for sel in selectable_group:
                if sel is not self and sel.selected:
                    self.selected = False
                    break
        return element


class Dragger:
    def __init__(
        self,
        position: typing.Sequence[float] | None = None,
        lock_x: bool = False,
        lock_y: bool = False,
    ):
        if position is None:
            position = (0, 0)
        self.position = pygame.Vector2(position)
        self.rel = pygame.Vector2()
        self.lock_x = lock_x
        self.lock_y = lock_y
        self.changed: bool = False
        self.clamp_x: tuple[int, int] | None = None
        self.clamp_y: tuple[int, int] | None = None
        self.style: _typing.ElementStyleLike = {"ignore_grid": True}
        self._before_update_pos = pygame.Vector2()

    def update[IT: _data.Interaction | _data.ElementData](
        self,
        element: IT,
    ) -> IT:
        if _core._globalctx._mili is None:
            raise _error.MILIStatusError("MILI.start() not called")
        self.changed = False
        self._before_update_pos = self.position.copy()
        if element.left_pressed:
            rel = _core._globalctx._mili._ctx._mouse_rel.copy()
            if self.lock_x:
                rel.x = 0
            if self.lock_y:
                rel.y = 0
            prev = self.position.copy()
            self.position += rel
            if self.position != prev:
                self.changed = True
            self.rel = rel.copy()
        return element

    def clamp(
        self,
        clamp_x: tuple[int, int] | None = None,
        clamp_y: tuple[int, int] | None = None,
        previous_clamps: bool = True,
    ) -> bool:
        if previous_clamps:
            clamp_x = self.clamp_x
            clamp_y = self.clamp_y
        prev = self.position.copy()
        if clamp_x is not None:
            self.position.x = pygame.math.clamp(self.position.x, *clamp_x)
        if clamp_y is not None:
            self.position.y = pygame.math.clamp(self.position.y, *clamp_y)
        if self.position != prev:
            self.changed = True
        if self.changed and self.position == self._before_update_pos:
            self.changed = False
        self.clamp_x = clamp_x
        self.clamp_y = clamp_y
        return self.changed


class GenericApp:
    def __init__(self, window: pygame.Window):
        if not pygame.get_init():
            raise _error.MILIStatusError(
                "pygame.init() must be called before creating a generic app"
            )
        self.window = window
        self.clock = pygame.Clock()
        self.target_framerate: int = 60
        self.mili = _MILI(self.window.get_surface())
        self.clear_color: _typing.ColorLike = 0
        self.start_style: _typing.ElementStyleLike | None = None
        self.delta_time: float = 0

    def update(self): ...

    def ui(self): ...

    def event(self, event: pygame.Event): ...

    def quit(self):
        self.on_quit()
        pygame.quit()
        raise SystemExit

    def on_quit(self): ...

    def run(self):
        while True:
            self.mili.start(self.start_style)
            for event in pygame.event.get():
                if event.type == pygame.WINDOWCLOSE and event.window == self.window:
                    self.quit()
                else:
                    self.event(event)

            self.window.get_surface().fill(self.clear_color)
            self.update()
            self.ui()
            self.mili.update_draw()
            self.window.flip()
            self.delta_time = self.clock.tick(self.target_framerate) / 1000


class Scroll:
    def __init__(self):
        self.scroll_offset: pygame.Vector2 = pygame.Vector2()
        self._element_data: _data.ElementData = None

    def update(self, element_data: _data.ElementData) -> _data.ElementData:
        if isinstance(element_data, _data.Interaction):
            raise _error.MILIError(
                "Scroll.update does not allow Interaction objects. Set get_data=True to get the correct ElementData object"
            )
        self._element_data = element_data
        self.clamp()
        return element_data

    def scroll(self, x: float, y: float):
        self.scroll_offset.x += x
        self.scroll_offset.y += y
        self.clamp()

    def set_scroll(self, x: float, y: float):
        self.scroll_offset.x = x
        self.scroll_offset.y = y
        self.clamp()

    def clamp(self):
        maxx, maxy = float("inf"), float("inf")
        if self._element_data is not None and self._element_data.grid is not None:
            maxx = self._element_data.grid.overflowx
            maxy = self._element_data.grid.overflowy
        self.scroll_offset.x = pygame.math.clamp(self.scroll_offset.x, 0, maxx)
        self.scroll_offset.y = pygame.math.clamp(self.scroll_offset.y, 0, maxy)

    def get_offset(self) -> tuple[float, float]:
        return (-self.scroll_offset.x, -self.scroll_offset.y)

    def get_handle_size(
        self, scrollbar_size: int, axis: typing.Literal["x", "y"] = "y"
    ):
        if self._element_data is not None and self._element_data.grid is not None:
            if axis == "x":
                rectsize = self._element_data.rect.w
                overflow = self._element_data.grid.overflowx
            else:
                rectsize = self._element_data.rect.h
                overflow = self._element_data.grid.overflowy
            return min(
                scrollbar_size, (rectsize / (rectsize + overflow)) * scrollbar_size
            )
        return 0

    def get_rel_handle_pos_from_scroll(
        self,
        scrollbar_size: int,
        handle_size: int,
        axis: typing.Literal["x", "y"] = "y",
    ):
        if self._element_data is None or self._element_data.grid is None:
            return 0
        if axis == "x":
            offset = self.scroll_offset.x
            overflow = self._element_data.grid.overflowx
        else:
            offset = self.scroll_offset.y
            overflow = self._element_data.grid.overflowy
        if overflow == 0:
            return 0
        return (offset / overflow) * (scrollbar_size - handle_size)

    def set_scroll_from_rel_handle_pos(
        self,
        scrollbar_size: int,
        rel_handle_pos: int,
        handle_size: int,
        axis: typing.Literal["x", "y"] = "y",
    ):
        if self._element_data is None or self._element_data.grid is None:
            return
        if scrollbar_size - handle_size == 0:
            return
        if axis == "x":
            self.scroll_offset.x = self._element_data.grid.overflowx * (
                rel_handle_pos / (scrollbar_size - handle_size)
            )
        else:
            self.scroll_offset.y = self._element_data.grid.overflowy * (
                rel_handle_pos / (scrollbar_size - handle_size)
            )
        self.clamp()


class Scrollbar:
    def __init__(
        self,
        scroll: Scroll,
        short_size: int,
        border_dist: int = 3,
        padding: int = 3,
        size_reduce: int = 0,
        axis: typing.Literal["x", "y"] = "y",
    ):
        self._scroll = scroll
        self.axis = axis
        self.short_size = short_size
        self.border_dist = border_dist
        self.padding = padding
        self.size_reduce = size_reduce
        self.size: int = 0
        self.handle_size: int = 0
        self.handle_dragger: Dragger = Dragger(
            lock_x=self.axis == "y", lock_y=self.axis == "x"
        )
        self.bar_rect: pygame.Rect = pygame.Rect()
        self.handle_rect: pygame.Rect = pygame.Rect()

        self.bar_style: _typing.ElementStyleLike = {"ignore_grid": True, "z": 999}
        self.handle_style: _typing.ElementStyleLike = {"ignore_grid": True, "z": 9999}

    def change_axis(self, axis: typing.Literal["x", "y"]):
        self.axis = axis
        self.handle_dragger.lock_x = self.axis == "x"
        self.handle_dragger.lock_y = self.axis == "y"

    @property
    def needed(self):
        if (
            self._scroll._element_data is None
            or self._scroll._element_data.grid is None
        ):
            return False
        return (
            self._scroll._element_data.grid.overflowy
            if self.axis == "y"
            else self._scroll._element_data.grid.overflowx
        ) > 0

    def update(self, element_data: _data.ElementData) -> _data.ElementData:
        if isinstance(element_data, _data.Interaction):
            raise _error.MILIError(
                "Scrollbar.update does not allow Interaction objects. Set get_data=True to get the correct ElementData object"
            )
        rectsize = element_data.rect.h if self.axis == "x" else element_data.rect.w
        previous = self.size
        self.size = (
            (element_data.rect.w if self.axis == "x" else element_data.rect.h)
            - self.padding * 2
            - self.size_reduce
        )
        if previous != self.size and previous != 0:
            inc = self.size / previous
            if self.axis == "x":
                self.handle_dragger.position.x *= inc
            else:
                self.handle_dragger.position.y *= inc
            self.handle_dragger.changed = True
        if self.axis == "x":
            self.bar_rect = pygame.Rect(
                self.padding,
                rectsize - self.border_dist - self.short_size,
                self.size,
                self.short_size,
            )
        else:
            self.bar_rect = pygame.Rect(
                rectsize - self.border_dist - self.short_size,
                self.padding,
                self.short_size,
                self.size,
            )
        self.handle_size = self._scroll.get_handle_size(self.size, self.axis)
        return element_data

    def update_handle[IT: _data.ElementData | _data.Interaction](
        self, element: IT
    ) -> IT:
        changed = self.handle_dragger.changed
        self.handle_dragger.update(
            element,
        )
        self.handle_dragger.clamp(
            clamp_x=(0, self.size - self.handle_size) if self.axis == "x" else None,
            clamp_y=(0, self.size - self.handle_size) if self.axis == "y" else None,
            previous_clamps=False,
        )
        if self.handle_dragger.changed or changed:
            self._scroll.set_scroll_from_rel_handle_pos(
                self.size,
                (
                    self.handle_dragger.position.x
                    if self.axis == "x"
                    else self.handle_dragger.position.y
                ),
                self.handle_size,
                self.axis,
            )
        if self.axis == "x":
            self.handle_rect = pygame.Rect(
                self.handle_dragger.position.x, 0, self.handle_size, self.short_size
            )
        else:
            self.handle_rect = pygame.Rect(
                0, self.handle_dragger.position.y, self.short_size, self.handle_size
            )

    def scroll_moved(self):
        pos = self._scroll.get_rel_handle_pos_from_scroll(
            self.size, self.handle_size, self.axis
        )
        if self.axis == "x":
            self.handle_dragger.position.x = pos
        else:
            self.handle_dragger.position.y = pos
        self.handle_dragger.clamp()


class Slider:
    def __init__(
        self,
        lock_x: bool,
        lock_y: bool,
        handle_size: typing.Sequence[float],
        strict_borders: bool = False,
    ):
        self.lock_x = lock_x
        self.lock_y = lock_y
        self.handle_size = handle_size
        self.strict_borders = strict_borders
        self._area_data: _data.ElementData | None = None
        self.handle_dragger: Dragger = Dragger()
        self.moved: bool = False

        self.area_style: _typing.ElementStyleLike = {"clip_draw": False}
        self.handle_style: _typing.ElementStyleLike = {"ignore_grid": True}
        self.handle_rect: pygame.Rect = pygame.Rect()

    @classmethod
    def from_axis(
        cls,
        axis: typing.Literal["x", "y"],
        handle_size: typing.Sequence[float],
        strict_borders: bool = False,
    ):
        return cls(axis == "y", axis == "x", handle_size, strict_borders)

    def update_area(self, element_data: _data.ElementData) -> _data.ElementData:
        if isinstance(element_data, _data.Interaction):
            raise _error.MILIError(
                "Slider.update_area does not allow Interaction objects. Set get_data=True to get the correct ElementData object"
            )
        previous = self.value
        self._area_data = element_data
        if self.value != previous:
            self.value = previous
        return element_data

    def update_handle[IT: _data.ElementData | _data.Interaction](
        self, element: IT
    ) -> IT:
        if self._area_data is None:
            raise _error.MILIStatusError(
                "Slider.update_area must be called before Slider.update_handle"
            )
        self.handle_dragger.lock_x = self.lock_x
        self.handle_dragger.lock_y = self.lock_y
        if self.lock_x:
            self.handle_dragger.position.x = self._area_data.rect.w / 2
        if self.lock_y:
            self.handle_dragger.position.y = self._area_data.rect.h / 2
        clamp_padx, clamp_pady = (
            self.handle_size[0] / 2 if self.strict_borders else 0,
            self.handle_size[1] / 2 if self.strict_borders else 0,
        )
        self.handle_dragger.update(
            element,
        )
        self.handle_dragger.clamp(
            (0 + clamp_padx, self._area_data.rect.w - clamp_padx),
            (0 + clamp_pady, self._area_data.rect.h - clamp_pady),
            False,
        )
        if self._area_data.rect.w < self.handle_size[0]:
            self.handle_dragger.position.x = (
                self.handle_size[0] / 2 if self.strict_borders else 0
            )
        if self._area_data.rect.h < self.handle_size[1]:
            self.handle_dragger.position.y = (
                self.handle_size[1] / 2 if self.strict_borders else 0
            )
        self.handle_rect = pygame.Rect(
            (
                self.handle_dragger.position.x - self.handle_size[0] / 2,
                self.handle_dragger.position.y - self.handle_size[1] / 2,
            ),
            self.handle_size,
        )
        self.moved = self.handle_dragger.changed
        return element

    @property
    def value(self) -> pygame.Vector2:
        if self._area_data is None:
            return pygame.Vector2(-1, -1)
        if not self.strict_borders:
            return pygame.Vector2(
                0
                if self.lock_x or self._area_data.rect.w == 0
                else self.handle_dragger.position.x / self._area_data.rect.w,
                0
                if self.lock_y or self._area_data.rect.h == 0
                else self.handle_dragger.position.y / self._area_data.rect.h,
            )
        hx, hy = self.handle_size[0] / 2, self.handle_size[1] / 2
        return pygame.Vector2(
            0
            if self.lock_x or self._area_data.rect.w == 0
            else (self.handle_dragger.position.x - hx)
            / (self._area_data.rect.w - hx * 2),
            0
            if self.lock_y or self._area_data.rect.h == 0
            else (self.handle_dragger.position.y - hy)
            / (self._area_data.rect.h - hy * 2),
        )

    @value.setter
    def value(self, v: typing.Sequence[float]):
        if self._area_data is None:
            return
        v = (pygame.math.clamp(v[0], 0, 1), pygame.math.clamp(v[1], 0, 1))
        if not self.strict_borders:
            if not self.lock_x:
                self.handle_dragger.position.x = v[0] * self._area_data.rect.w
            if not self.lock_y:
                self.handle_dragger.position.y = v[1] * self._area_data.rect.h
            self.handle_rect = pygame.Rect(
                (
                    self.handle_dragger.position.x - self.handle_size[0] / 2,
                    self.handle_dragger.position.y - self.handle_size[1] / 2,
                ),
                self.handle_size,
            )
            return
        hx, hy = self.handle_size[0] / 2, self.handle_size[1] / 2
        if not self.lock_x:
            self.handle_dragger.position.x = (
                v[0] * (self._area_data.rect.w - hx * 2) + hx
            )
        if not self.lock_y:
            self.handle_dragger.position.y = (
                v[1] * (self._area_data.rect.h - hy * 2) + hy
            )
        self.handle_rect = pygame.Rect(
            (
                self.handle_dragger.position.x - self.handle_size[0] / 2,
                self.handle_dragger.position.y - self.handle_size[1] / 2,
            ),
            self.handle_size,
        )

    @property
    def valuex(self) -> float:
        return self.value.x

    @property
    def valuey(self) -> float:
        return self.value.y

    @valuex.setter
    def valuex(self, v: float):
        self.value = pygame.Vector2(v, self.value.y)

    @valuey.setter
    def valuey(self, v: float):
        self.value = pygame.Vector2(self.value.x, v)


class InteractionSound:
    def __init__(
        self,
        hover: pygame.mixer.Sound | None = None,
        unhover: pygame.mixer.Sound | None = None,
        press: pygame.mixer.Sound | None = None,
        release: pygame.mixer.Sound | None = None,
        button: int = pygame.BUTTON_LEFT,
    ):
        self.hover_sound = hover
        self.unhover_sound = unhover
        self.press_sound = press
        self.release_sound = release
        self.button = button

    def play[IT: _data.Interaction | _data.ElementData](
        self,
        interaction: IT,
        channel: pygame.Channel | None = None,
        maxtime: int = 0,
        fade_ms: int = 0,
    ) -> IT:
        if interaction.just_hovered and self.hover_sound is not None:
            if channel is not None:
                channel.play(self.hover_sound, 0, maxtime, fade_ms)
            else:
                self.hover_sound.play(0, maxtime, fade_ms)
        if interaction.just_unhovered and self.unhover_sound is not None:
            if channel is not None:
                channel.play(self.unhover_sound, 0, maxtime, fade_ms)
            else:
                self.unhover_sound.play(0, maxtime, fade_ms)
        if (
            interaction.just_pressed_button == self.button
            and self.press_sound is not None
        ):
            if channel is not None:
                channel.play(self.press_sound, 0, maxtime, fade_ms)
            else:
                self.press_sound.play(0, maxtime, fade_ms)
        if (
            interaction.just_released_button == self.button
            and self.release_sound is not None
        ):
            if channel is not None:
                channel.play(self.release_sound, 0, maxtime, fade_ms)
            else:
                self.release_sound.play(0, maxtime, fade_ms)
        return interaction


class CustomWindowBorders:
    def __init__(
        self,
        window: pygame.Window,
        border_size: int = 3,
        corner_size: int = 6,
        titlebar_height: int = 30,
        change_cursor: bool = True,
        minimum_ratio: float | None = None,
        ratio_axis: typing.Literal["x", "y"] = "y",
        uniform_resize_key: int | None = pygame.K_LSHIFT,
        allowed_directions: list[
            typing.Literal[
                "top",
                "left",
                "bottom",
                "right",
                "topleft",
                "topright",
                "bottomleft",
                "bottomright",
            ]
        ]
        | typing.Literal["all"] = "all",
        on_move: typing.Callable[[], None] | None = None,
        on_resize: typing.Callable[[], None] | None = None,
        on_start_move: typing.Callable[[], None] | None = None,
        on_start_resize: typing.Callable[[], None] | None = None,
        on_end_move: typing.Callable[[], None] | None = None,
        on_end_resize: typing.Callable[[], None] | None = None,
    ):
        self.window = window
        self.border_size = border_size
        self.corner_size = corner_size
        self.titlebar_height = titlebar_height
        self.change_cursor = change_cursor
        self.minimum_ratio = minimum_ratio
        self.ratio_axis = ratio_axis
        self.uniform_resize_key = uniform_resize_key
        self.allowed_directions = allowed_directions
        self._callbacks: dict[
            typing.Literal["move", "resize"],
            dict[typing.Literal["start", "end", "during"]],
            typing.Callable[[], None] | None,
        ] = {
            "move": {"start": on_start_move, "end": on_end_move, "during": on_move},
            "resize": {
                "start": on_start_resize,
                "end": on_end_resize,
                "during": on_resize,
            },
        }
        self.dragging: bool = False
        self.resizing: bool = False
        self.relative: pygame.Vector2 = pygame.Vector2()
        self.cumulative_relative: pygame.Vector2 = pygame.Vector2()

        self._press_rel = pygame.Vector2()
        self._press_global = pygame.Vector2()
        self._start_val = pygame.Vector2()
        self._resize_winsize = None
        self._resize_winpos = None
        self._resize_handle = None
        self._resize_handles = [
            CustomWindowBorders._ResizeHandle(
                self, "topleft", True, None, "xy", pygame.SYSTEM_CURSOR_SIZENWSE
            ),
            CustomWindowBorders._ResizeHandle(
                self, "topright", True, None, "y", pygame.SYSTEM_CURSOR_SIZENESW
            ),
            CustomWindowBorders._ResizeHandle(
                self, "bottomleft", True, None, "x", pygame.SYSTEM_CURSOR_SIZENESW
            ),
            CustomWindowBorders._ResizeHandle(
                self, "bottomright", True, None, None, pygame.SYSTEM_CURSOR_SIZENWSE
            ),
            CustomWindowBorders._ResizeHandle(
                self, "top", False, "x", "y", pygame.SYSTEM_CURSOR_SIZENS
            ),
            CustomWindowBorders._ResizeHandle(
                self, "left", False, "y", "x", pygame.SYSTEM_CURSOR_SIZEWE
            ),
            CustomWindowBorders._ResizeHandle(
                self, "bottom", False, "x", None, pygame.SYSTEM_CURSOR_SIZENS
            ),
            CustomWindowBorders._ResizeHandle(
                self, "right", False, "y", None, pygame.SYSTEM_CURSOR_SIZEWE
            ),
        ]

    def update(self) -> bool:
        just = pygame.mouse.get_just_pressed()[0]
        mpos = pygame.Vector2(pygame.mouse.get_pos())
        pressed = pygame.mouse.get_pressed()[0]
        modified_cursor = False

        tbar_rect = pygame.Rect(0, 0, self.window.size[0], self.titlebar_height)
        for handle in self._resize_handles:
            if (
                self.allowed_directions == "all"
                or handle._name in self.allowed_directions
            ):
                handle._make_rect()

        if self.change_cursor:
            if not self.resizing:
                if tbar_rect.h > 0 and (tbar_rect.collidepoint(mpos) or self.dragging):
                    pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_HAND)
                    modified_cursor = True
            for handle in self._resize_handles:
                if (
                    self.allowed_directions == "all"
                    or handle._name in self.allowed_directions
                ) and (
                    handle._rect.collidepoint(mpos)
                    or (self.resizing and self._resize_handle is handle)
                ):
                    pygame.mouse.set_cursor(handle._cursor)
                    modified_cursor = True
                    break

        if just:
            for handle in self._resize_handles:
                if (
                    self.allowed_directions == "all"
                    or handle._name in self.allowed_directions
                ) and handle._rect.collidepoint(mpos):
                    self.resizing = True
                    self._resize_handle = handle
                    self._press_global = mpos + self.window.position
                    self._press_rel = mpos
                    self._resize_winpos = self.window.position
                    self._resize_winsize = self.window.size
                    self._start_val = pygame.Vector2(self.window.size)
                    callback = self._callbacks["resize"]["start"]
                    if callback is not None:
                        callback()
                    break

        if pressed:
            if self.resizing:
                self._resize_handle._update(mpos)
        else:
            if self.resizing:
                self.cumulative_relative = self.window.size - self._start_val
                callback = self._callbacks["resize"]["end"]
                if callback is not None:
                    callback()
            self.resizing = False
            self._resize_handle = None
            self.relative = pygame.Vector2()

        if self.resizing or tbar_rect.h <= 0:
            return modified_cursor

        if just and tbar_rect.collidepoint(mpos):
            self._press_rel = mpos
            self._press_global = mpos + self.window.position
            self._start_val = pygame.Vector2(self.window.position)
            self.dragging = True
            callback = self._callbacks["move"]["start"]
            if callback is not None:
                callback()

        if pressed:
            if not just and self.dragging:
                previous = pygame.Vector2(self.window.position)
                new = previous + pygame.mouse.get_pos()
                self.window.position = (
                    self._press_global + (new - self._press_global) - self._press_rel
                )
                self.relative = self.window.position - previous
                self.cumulative_relative = self.window.position - self._start_val
                callback = self._callbacks["move"]["during"]
                if callback is not None:
                    callback()
        else:
            if self.dragging:
                self.cumulative_relative = self.window.position - self._start_val
                callback = self._callbacks["move"]["end"]
                if callback is not None:
                    callback()
            self.dragging = False
            self.relative = pygame.Vector2()
        return modified_cursor

    class _ResizeHandle:
        def __init__(
            self,
            parent: "CustomWindowBorders",
            name,
            iscorner,
            axis,
            movewindow,
            cursor,
        ):
            self._parent = parent
            self._name = name
            self._iscorner = iscorner
            self._axis = axis
            self._movewindow = movewindow
            self._cursor = cursor
            self._axis_lock = (
                "x" if self._axis == "y" else "y" if self._axis == "x" else None
            )
            self._rect: pygame.Rect = pygame.Rect()

        def _make_rect(self):
            if self._iscorner:
                rect = pygame.Rect(
                    0, 0, self._parent.corner_size, self._parent.corner_size
                )
            else:
                rect = pygame.Rect(
                    0,
                    0,
                    self._parent.window.size[0]
                    if self._axis == "x"
                    else self._parent.border_size,
                    self._parent.window.size[1]
                    if self._axis == "y"
                    else self._parent.border_size,
                )
            if self._name == "topright":
                rect = rect.move_to(topright=(self._parent.window.size[0], 0))
            elif self._name == "bottomleft" or self._name == "bottom":
                rect = rect.move_to(bottomleft=(0, self._parent.window.size[1]))
            elif self._name == "bottomright" or self._name == "right":
                rect = rect.move_to(bottomright=self._parent.window.size)
            self._rect = rect

        def _update(self, mpos):
            previous = pygame.Vector2(self._parent.window.size)
            rel: pygame.Vector2 = (
                self._parent.window.position + mpos - self._parent._press_global
            )
            posrel = pygame.Vector2()
            if self._axis_lock == "y":
                rel.x = 0
            elif self._axis_lock == "x":
                rel.y = 0
            if self._movewindow == "x":
                rel.x *= -1
            elif self._movewindow == "y":
                rel.y *= -1
            elif self._movewindow == "xy":
                rel.x *= -1
                rel.y *= -1
            if (
                self._parent.uniform_resize_key is not None
                and pygame.key.get_pressed()[self._parent.uniform_resize_key]
            ):
                if rel.x != rel.y:
                    if abs(rel.x) > abs(rel.y):
                        rel.y = rel.x
                    else:
                        rel.x = rel.y
            if self._movewindow == "x":
                posrel.x = rel.x
            elif self._movewindow == "y":
                posrel.y = rel.y
            elif self._movewindow == "xy":
                posrel = rel
            newsize = self._parent._resize_winsize + rel
            if newsize.x <= 0 or newsize.y <= 0:
                return
            if self._parent.minimum_ratio is not None:
                if self._parent.ratio_axis:
                    ratio = newsize[0] / newsize[1]
                else:
                    ratio = newsize[1] / newsize[0]
                if ratio < self._parent.minimum_ratio:
                    if self._parent.ratio_axis == "y":
                        newsize = pygame.Vector2(
                            newsize[1] * self._parent.minimum_ratio, newsize[1]
                        )
                    else:
                        newsize = pygame.Vector2(
                            newsize[0], newsize[0] * self._parent.minimum_ratio
                        )
            self._parent.window.size = newsize
            diff = self._parent.window.size - newsize
            if self._movewindow == "x" and diff.x != 0:
                posrel.x += diff.x
            elif self._movewindow == "y" and diff.y != 0:
                posrel.y += diff.y
            elif self._movewindow == "xy" and diff.length() != 0:
                posrel += diff
            self._parent.relative = self._parent.window.size - previous
            self._parent.cumulative_relative = (
                self._parent.window.size - self._parent._start_val
            )
            if posrel.length() != 0:
                self._parent.window.position = self._parent._resize_winpos - posrel
            callback = self._parent._callbacks["resize"]["during"]
            if callback is not None:
                callback()


def percentage(percentage: float, value: float) -> float:
    return (percentage * value) / 100


def indent(*args, **kwargs): ...


def fit_image(
    rect: _typing.RectLike,
    surface: pygame.Surface,
    padx: int = 0,
    pady: int = 0,
    do_fill: bool = False,
    do_stretchx: bool = False,
    do_stretchy: bool = False,
    fill_color: _typing.ColorLike | None = None,
    border_radius: int = 0,
    alpha: int = 255,
    smoothscale: bool = False,
    ninepatch: int = 0,
) -> pygame.Surface:
    return _core._globalctx._get_image(
        pygame.Rect(rect),
        surface,
        int(padx),
        int(pady),
        do_fill,
        do_stretchx,
        do_stretchy,
        fill_color,
        int(border_radius),
        int(alpha),
        smoothscale,
        int(ninepatch),
    )
